require 'upvote/version'
require 'upvote/configuration'

require 'active_support/core_ext/string/inflections'
require 'active_support/core_ext/object/blank'
require 'redis'

module Upvote
  extend self
  extend Configuration

  def create(subject, target, execute: true)
    creation = ->() {
      redis.zadd set_key(subject), Time.now.to_f, key(target)
      redis.zadd set_key(target),  Time.now.to_f, key(subject)
    }

    execute ? redis.multi { creation.call } : creation
  end

  def remove(subject, target, execute: true)
    removal = ->() {
      redis.zrem set_key(subject), key(target)
      redis.zrem set_key(target),  key(subject)
    }

    execute ? redis.multi { removal.call } : removal
  end

  def exist?(subject, target)
    redis.zscore(set_key(target), key(subject)).present?
  end

  def count(entity)
    redis.zcard set_key(entity)
  end

  def voters(entity, start: 0, offset: -1)
    redis.zrevrange set_key(entity), start, offset
  end

  def multi_create(upvotes)
    redis.multi do
      upvotes.each do |subject, *targets|
        targets.each do |target|
          create(subject, target, execute: false).call
        end
      end
    end
  end

  def multi_remove(removals)
    redis.multi do
      removals.each do |subject, *targets|
        targets.each do |target|
          remove(subject, target, execute: false).call
        end
      end
    end
  end

  def multi_count(entities)
    redis.multi { entities.each &method(:count) }
  end

  def multi_voters(entities, start: 0, offset: -1)
    redis.multi {
      entities.each { |e| voters entities, start: start, offset: offset }
    }
  end

  private

  def set_key(entity)
    "#{key(entity)}:likes"
  end

  def key(entity)
    klass  = entity.class
    method = id_method(klass)
    '${type}#${id}' % { type: klass.to_s.underscore, id: entity.send(method) }
  end

end
